from __future__ import annotations

import codecs
import ipaddress
import secrets
import string
import time
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from ipaddress import IPv4Network, IPv6Network
from typing import TYPE_CHECKING, Any

from botocore.exceptions import ClientError
from pydantic import BaseModel

from bt_ddos_shield.utils import ShieldAddress

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence
    from types import MappingProxyType

    from mypy_boto3_ec2 import EC2Client
    from mypy_boto3_ec2.type_defs import (
        CreateSecurityGroupResultTypeDef,
        CreateSubnetResultTypeDef,
        CreateVpcResultTypeDef,
        DescribeAvailabilityZonesResultTypeDef,
        DescribeInstancesResultTypeDef,
        DescribeSubnetsResultTypeDef,
        DescribeVpcsResultTypeDef,
        GroupIdentifierTypeDef,
        InstanceTypeDef,
        SubnetTypeDef,
        VpcTypeDef,
    )
    from mypy_boto3_elbv2 import ElasticLoadBalancingv2Client
    from mypy_boto3_elbv2.type_defs import (
        CreateLoadBalancerOutputTypeDef,
        CreateTargetGroupOutputTypeDef,
        DescribeLoadBalancersOutputTypeDef,
        LoadBalancerTypeDef,
    )
    from mypy_boto3_route53 import Route53Client
    from mypy_boto3_route53.type_defs import (
        ListResourceRecordSetsResponseTypeDef,
        ResourceRecordSetOutputTypeDef,
        ResourceRecordSetTypeDef,
    )
    from mypy_boto3_wafv2 import WAFV2Client
    from mypy_boto3_wafv2.type_defs import (
        CreateWebACLResponseTypeDef,
        GetWebACLResponseTypeDef,
        RuleOutputTypeDef,
        RuleTypeDef,
        WebACLTypeDef,
    )
    from route53.connection import Route53Connection
    from route53.hosted_zone import HostedZone
    from route53.resource_record_set import ResourceRecordSet

    from bt_ddos_shield.event_processor import AbstractMinerShieldEventProcessor
    from bt_ddos_shield.state_manager import AbstractMinerShieldStateManager, MinerShieldState
    from bt_ddos_shield.utils import AWSClientFactory, Hotkey


class ShieldedServerLocationType(Enum):
    """
    Possible types of shielded server location.
    """

    EC2_ID = 'ec2_id'
    """ ID of EC2 instance """
    EC2_IP = 'ec2_ip'
    """ IPv4 address of EC2 instance """


@dataclass
class ShieldedServerLocation:
    """
    Location of server, which shield should protect.
    """

    location_type: ShieldedServerLocationType
    location_value: str
    """ Value depends on location type """
    port: int
    """ Port used by shielded server """


class AddressManagerException(Exception):
    pass


class AbstractAddressManager(ABC):
    """
    Abstract base class for manager handling public IP/domain addresses assigned to validators.
    """

    def hide_original_server(self):  # noqa: B027
        """
        If method is implemented, it should hide the original server IP address from public access.
        See auto_hide_original_server in MinerShield options.
        """
        pass

    @abstractmethod
    def clean_all(self):
        """
        Clean everything created before by the address manager.
        """
        pass

    @abstractmethod
    def create_address(self, hotkey: Hotkey) -> ShieldAddress:
        """
        Create and return a new address redirecting to Miner server to be used by validator identified by hotkey.
        """
        pass

    @abstractmethod
    def remove_address(self, address: ShieldAddress):
        pass

    @abstractmethod
    def validate_addresses(self, addresses: MappingProxyType[Hotkey, ShieldAddress]) -> set[Hotkey]:
        """
        Validate if given addresses exist and are working properly.

        Args:
            addresses: Dictionary of addresses to validate (validator HotKey -> Address).

        Returns:
            set[Hotkey]: Set of HotKeys of validators with invalid addresses.
        """
        pass


@dataclass
class AwsEC2InstanceData:
    instance_id: str
    vpc_id: str
    subnet_id: str
    private_ip: str
    security_groups: list[GroupIdentifierTypeDef]


@dataclass
class AwsSubnetData:
    subnet_id: str
    availability_zone: str
    cidr_block: str


@dataclass
class AwsVpcData:
    vpc_id: str
    cidr_block: str
    subnets: list[AwsSubnetData]


@dataclass
class AwsELBData:
    id: str
    dns_name: str
    hosted_zone_id: str


class AwsEC2ServerLocation(BaseModel):
    vpc_id: str
    subnet_id: str
    server_id: str


class AwsShieldedServerData(BaseModel):
    server_location: ShieldedServerLocation
    """ Location of shielded server. """
    aws_location: AwsEC2ServerLocation | None
    """ Detailed location of server in AWS (only if it is EC2 instance). """

    def to_json(self) -> str:
        return self.model_dump_json()

    @staticmethod
    def from_json(json_str: str) -> AwsShieldedServerData:
        return AwsShieldedServerData.model_validate_json(json_str)


class AwsObjectTypes(Enum):
    WAF = 'WAF'
    ELB = 'ELB'
    SUBNET = 'SUBNET'
    VPC = 'VPC'
    DNS_ENTRY = 'DNS_ENTRY'
    TARGET_GROUP = 'TARGET_GROUP'
    SECURITY_GROUP = 'SECURITY_GROUP'


class AwsAddressManager(AbstractAddressManager):
    """
    Address manager using AWS Route53 service to manage DNS records and ELB for handling access to Miner server.
    """

    shielded_server_data: AwsShieldedServerData
    waf_client: WAFV2Client
    waf_arn: str | None
    elb_client: ElasticLoadBalancingv2Client
    elb_data: AwsELBData | None
    ec2_client: EC2Client
    hosted_zone_id: str
    """ ID of hosted zone in Route53 where addresses are located. """
    hosted_zone: HostedZone
    route53_client: Route53Connection
    route53_boto_client: Route53Client
    event_processor: AbstractMinerShieldEventProcessor
    state_manager: AbstractMinerShieldStateManager

    HOSTED_ZONE_ID_STATE_KEY: str = 'aws_hosted_zone_id'
    SHIELDED_SERVER_STATE_KEY: str = 'aws_shielded_server'

    AWS_OPERATION_MAX_RETRIES: int = 20
    """ How many retries should be done for AWS operations, which need long time until can be processed. """
    AWS_OPERATION_RETRY_DELAY_SEC: int = 9
    """
    Delay in seconds between retries for AWS operations. Total time for waiting for operation is
    AWS_OPERATION_MAX_RETRIES * AWS_OPERATION_RETRY_DELAY_SEC seconds.
    """

    def __init__(
        self,
        aws_client_factory: AWSClientFactory,
        server_location: ShieldedServerLocation,
        hosted_zone_id: str,
        event_processor: AbstractMinerShieldEventProcessor,
        state_manager: AbstractMinerShieldStateManager,
    ):
        """
        Initialize AWS address manager.
        """
        self.event_processor = event_processor
        self.state_manager = state_manager

        self.waf_client = aws_client_factory.boto3_client('wafv2')  # type: ignore
        self.waf_arn = None
        self.elb_client = aws_client_factory.boto3_client('elbv2')  # type: ignore
        self.elb_data = None
        self.route53_client = aws_client_factory.route53_client()
        self.hosted_zone_id = hosted_zone_id
        self.hosted_zone = self.route53_client.get_hosted_zone_by_id(hosted_zone_id)
        self.route53_boto_client = aws_client_factory.boto3_client('route53')  # type: ignore
        self.ec2_client = aws_client_factory.boto3_client('ec2')  # type: ignore
        self._initialize_server_data(server_location)

    def _initialize_server_data(self, server_location: ShieldedServerLocation):
        server_instance: AwsEC2InstanceData
        if server_location.location_type == ShieldedServerLocationType.EC2_IP:
            server_instance = self._get_ec2_instance_data(private_ip=server_location.location_value)
        else:
            assert server_location.location_type == ShieldedServerLocationType.EC2_ID
            server_instance = self._get_ec2_instance_data(instance_id=server_location.location_value)

        server_location = ShieldedServerLocation(
            location_type=ShieldedServerLocationType.EC2_ID,
            location_value=server_instance.instance_id,
            port=server_location.port,
        )
        server_aws_location = AwsEC2ServerLocation(
            vpc_id=server_instance.vpc_id, subnet_id=server_instance.subnet_id, server_id=server_instance.instance_id
        )
        self.shielded_server_data = AwsShieldedServerData(
            server_location=server_location, aws_location=server_aws_location
        )

    def clean_all(self) -> None:
        created_objects: MappingProxyType[str, frozenset[str]] = (
            self.state_manager.get_state().address_manager_created_objects
        )

        self._delete_route53_records(self.hosted_zone_id)

        # Order of removal is important
        cleaned: bool = True
        cleaned = self._clean_aws_objects(created_objects, AwsObjectTypes.WAF, self._remove_firewall) and cleaned
        cleaned = self._clean_aws_objects(created_objects, AwsObjectTypes.ELB, self._remove_elb) and cleaned
        cleaned = (
            self._clean_aws_objects(created_objects, AwsObjectTypes.SECURITY_GROUP, self._remove_security_group)
            and cleaned
        )
        cleaned = (
            self._clean_aws_objects(created_objects, AwsObjectTypes.TARGET_GROUP, self._remove_target_group) and cleaned
        )
        cleaned = self._clean_aws_objects(created_objects, AwsObjectTypes.SUBNET, self._remove_subnet) and cleaned
        cleaned = self._clean_aws_objects(created_objects, AwsObjectTypes.VPC, self._remove_vpc) and cleaned
        if not cleaned:
            raise AddressManagerException(
                'Failed to clean all AWS objects that are no longer needed. Check logs to see what is left.'
                ' Also try to run cleaning later. Sometimes AWS needs 15-20 minutes to be able to clean some of'
                ' them - especially TargetGroups.'
            )

    @classmethod
    def _clean_aws_objects(
        cls,
        created_objects: MappingProxyType[str, frozenset[str]],
        object_type: AwsObjectTypes,
        remove_method: Callable[[str], bool],
    ) -> bool:
        if object_type.value not in created_objects:
            return True
        cleaned: bool = True
        for object_id in created_objects[object_type.value]:
            try:
                cleaned = remove_method(object_id) and cleaned
            except Exception as e:
                cls.event_processor.event(
                    'Failed to remove {object_type} AWS object with id={id}',
                    exception=e,
                    object_type=object_type.value,
                    id=object_id,
                )
                cleaned = False
        return cleaned

    def create_address(self, hotkey: Hotkey) -> ShieldAddress:
        self._validate_manager_state()

        subdomain: str = self._generate_subdomain(hotkey)
        new_address_domain: str = f'{subdomain}.{self._get_hosted_zone_domain(self.hosted_zone)}'
        assert self.waf_arn is not None, '_validate_manager_state creates WAF and should be called before'
        self._add_domain_rule_to_firewall(self.waf_arn, new_address_domain)
        return ShieldAddress(
            address_id=subdomain,
            address=new_address_domain,
            port=self.shielded_server_data.server_location.port,
        )

    @classmethod
    def _generate_subdomain(cls, hotkey: Hotkey) -> str:
        return f'{hotkey[:8]}_{secrets.token_urlsafe(16)}'.lower()

    @classmethod
    def _get_hosted_zone_domain(cls, hosted_zone: HostedZone) -> str:
        return hosted_zone.name[:-1]  # Cut '.' from the end of hosted zone name

    def remove_address(self, address: ShieldAddress):
        self._validate_manager_state()
        assert self.waf_arn is not None, '_validate_manager_state creates WAF and should be called before'
        self._remove_domain_rule_from_firewall(self.waf_arn, address.address)

    def validate_addresses(self, addresses: MappingProxyType[Hotkey, ShieldAddress]) -> set[Hotkey]:
        if self._validate_manager_state():
            return {hotkey for hotkey, _ in addresses.items()}

        if not addresses:
            return set()

        assert self.waf_arn is not None, '_validate_manager_state creates WAF and should be called before'
        waf_data: GetWebACLResponseTypeDef = self._get_firewall_info(self.waf_arn)
        rules: list[RuleOutputTypeDef] = waf_data['WebACL']['Rules']

        invalid_hotkeys: set[Hotkey] = set()
        for hotkey, address in addresses.items():
            rule: RuleOutputTypeDef | None = self._find_rule(rules, address.address)
            if rule is None:
                invalid_hotkeys.add(hotkey)
        return invalid_hotkeys

    def _validate_manager_state(self) -> bool:
        """Returns if we should invalidate all addresses created before."""
        ret: bool = self._handle_shielded_server_change()
        self.elb_data = self._create_elb_if_needed(self.shielded_server_data)
        self.waf_arn = self._create_firewall_if_needed()

        ret = self._handle_hosted_zone_change() or ret
        self._init_hosted_zone_if_needed()

        return ret

    def _store_server_data(self, server_data: AwsShieldedServerData):
        self.state_manager.update_address_manager_state(self.SHIELDED_SERVER_STATE_KEY, server_data.to_json())

    def _load_server_data(self) -> AwsShieldedServerData | None:
        state: MinerShieldState = self.state_manager.get_state()
        if self.SHIELDED_SERVER_STATE_KEY not in state.address_manager_state:
            return None
        json_data: str = state.address_manager_state[self.SHIELDED_SERVER_STATE_KEY]
        return AwsShieldedServerData.from_json(json_data)

    def _handle_shielded_server_change(self) -> bool:
        old_server_data: AwsShieldedServerData | None = self._load_server_data()
        if old_server_data != self.shielded_server_data:
            # If shielded server changed, we need to recreate ELB. Maybe we can try to change only
            # needed objects, but changing ELB is the easiest way and this operation should happen rarely.
            self.event_processor.event(
                'Shielded server changed from {old_server_desc} to {new_server_desc}',
                old_server_desc=str(old_server_data) if old_server_data else 'None',
                new_server_desc=str(self.shielded_server_data),
            )
            self.clean_all()
            self._store_server_data(self.shielded_server_data)
            return True

        return False

    def _handle_hosted_zone_change(self) -> bool:
        state: MinerShieldState = self.state_manager.get_state()
        zone_changed: bool = False
        if self.HOSTED_ZONE_ID_STATE_KEY in state.address_manager_state:
            old_zone_id: str = state.address_manager_state[self.HOSTED_ZONE_ID_STATE_KEY]
            if old_zone_id != self.hosted_zone_id:
                # If hosted zone changed, we need to clean all previous route53 addresses and WAF rules
                self.event_processor.event(
                    'Route53 hosted zone changed from {old_id} to {new_id}',
                    old_id=old_zone_id,
                    new_id=self.hosted_zone_id,
                )
                self._delete_route53_records(old_zone_id)
                assert self.waf_arn is not None, '_validate_manager_state creates WAF and should be called before'
                self._clear_firewall_rules(self.waf_arn)
                zone_changed = True
        else:
            zone_changed = True

        if zone_changed:
            self.state_manager.update_address_manager_state(self.HOSTED_ZONE_ID_STATE_KEY, self.hosted_zone_id)
        return zone_changed

    def _get_ec2_instance_data(
        self, instance_id: str | None = None, private_ip: str | None = None
    ) -> AwsEC2InstanceData:
        assert instance_id or private_ip
        ec2_client_args: dict[str, Any] = (
            {'InstanceIds': [instance_id]}
            if instance_id
            else {'Filters': [{'Name': 'private-ip-address', 'Values': [private_ip]}]}
        )
        response: DescribeInstancesResultTypeDef = self.ec2_client.describe_instances(**ec2_client_args)
        if not response['Reservations']:
            raise AddressManagerException(
                f'No EC2 instance found with instance_id={instance_id} or IP address {private_ip}'
            )
        instance_data: InstanceTypeDef = response['Reservations'][0]['Instances'][0]
        return AwsEC2InstanceData(
            instance_id=instance_data['InstanceId'],
            vpc_id=instance_data['VpcId'],
            subnet_id=instance_data['SubnetId'],
            private_ip=instance_data['PrivateIpAddress'],
            security_groups=instance_data['SecurityGroups'],
        )

    def _get_vpc_data(self, vpc_id: str) -> AwsVpcData:
        vpc_response: DescribeVpcsResultTypeDef = self.ec2_client.describe_vpcs(VpcIds=[vpc_id])
        vpc_data: VpcTypeDef = vpc_response['Vpcs'][0]
        cidr_block: str = vpc_data['CidrBlock']

        subnet_response: DescribeSubnetsResultTypeDef = self.ec2_client.describe_subnets(
            Filters=[{'Name': 'vpc-id', 'Values': [vpc_id]}]
        )
        subnets: list[AwsSubnetData] = [
            AwsSubnetData(
                subnet_id=subnet['SubnetId'],
                availability_zone=subnet['AvailabilityZone'],
                cidr_block=subnet['CidrBlock'],
            )
            for subnet in subnet_response['Subnets']
        ]
        return AwsVpcData(vpc_id=vpc_id, cidr_block=cidr_block, subnets=subnets)

    def _get_subnet_data(self, subnet_id: str) -> AwsSubnetData:
        response: DescribeSubnetsResultTypeDef = self.ec2_client.describe_subnets(SubnetIds=[subnet_id])
        subnet_data: SubnetTypeDef = response['Subnets'][0]
        return AwsSubnetData(
            subnet_id=subnet_id, availability_zone=subnet_data['AvailabilityZone'], cidr_block=subnet_data['CidrBlock']
        )

    def _add_route53_record(self, subdomain: str, hosted_zone: HostedZone):
        domain_name: str = f'{subdomain}.{hosted_zone.name}'
        assert self.elb_data is not None, '_validate_manager_state creates ELB and should be called before'
        record_set: ResourceRecordSetTypeDef = {
            'Name': domain_name,
            'Type': 'A',
            'AliasTarget': {
                'HostedZoneId': self.elb_data.hosted_zone_id,
                'DNSName': self.elb_data.dns_name,
                'EvaluateTargetHealth': False,
            },
        }

        # Route53Connection doesn't handle alias records properly, so we use boto3 client directly
        self.route53_boto_client.change_resource_record_sets(
            HostedZoneId=hosted_zone.id,
            ChangeBatch={
                'Changes': [
                    {
                        'Action': 'CREATE',
                        'ResourceRecordSet': record_set,
                    }
                ]
            },
        )
        self.event_processor.event(
            'Added Route53 record {domain_name} to hosted zone {zone_id}',
            domain_name=domain_name,
            zone_id=hosted_zone.id,
        )
        try:
            # There is no ID for Route53 addresses, so we use domain name as an ID
            self.state_manager.add_address_manager_created_object(AwsObjectTypes.DNS_ENTRY.value, domain_name)
        except Exception as e:
            self._delete_route53_record_by_domain_name(domain_name, hosted_zone)
            raise e

    def _delete_route53_record_by_domain_name(self, domain_name: str, hosted_zone: HostedZone):
        for record_set in hosted_zone.record_sets:
            if codecs.decode(record_set.name, 'unicode_escape') == domain_name:
                self._delete_route53_record(record_set, hosted_zone)
                return

    def _delete_route53_record(self, record_set: ResourceRecordSet, hosted_zone: HostedZone):
        # Route53Connection doesn't handle alias records properly, so we use boto3 client directly
        response: ListResourceRecordSetsResponseTypeDef = self.route53_boto_client.list_resource_record_sets(
            HostedZoneId=hosted_zone.id,
            StartRecordName=record_set.name,
            StartRecordType=record_set.rrset_type,
            MaxItems='1',
        )
        record_set_data: ResourceRecordSetOutputTypeDef = response['ResourceRecordSets'][0]

        self.route53_boto_client.change_resource_record_sets(
            HostedZoneId=hosted_zone.id,
            ChangeBatch={'Changes': [{'Action': 'DELETE', 'ResourceRecordSet': record_set_data}]},
        )
        self.event_processor.event(
            'Deleted Route53 record {name} from hosted zone {zone_id}', name=record_set.name, zone_id=hosted_zone.id
        )
        self.state_manager.del_address_manager_created_object(AwsObjectTypes.DNS_ENTRY.value, record_set.name)

    @classmethod
    def _generate_random_alnum_string(cls, length: int) -> str:
        characters = string.ascii_letters + string.digits
        return ''.join(secrets.choice(characters) for _ in range(length))

    def _get_vpc_networks(self) -> list[IPv4Network | IPv6Network]:
        response: DescribeVpcsResultTypeDef = self.ec2_client.describe_vpcs()
        return [ipaddress.ip_network(vpc['CidrBlock']) for vpc in response['Vpcs']]

    @classmethod
    def _get_subnet_networks(cls, vpc: AwsVpcData) -> list[IPv4Network | IPv6Network]:
        return [ipaddress.ip_network(subnet.cidr_block) for subnet in vpc.subnets]

    @classmethod
    def _find_available_subnet(
        cls, network: IPv4Network | IPv6Network, used_subnets: list[IPv4Network | IPv6Network], subnet_size: int
    ) -> str:
        """
        Find available CIDR block (subnet) of specified size inside network. This block must not collide with any
        subnets from used_subnets param.
        """
        for subnet in network.subnets(new_prefix=subnet_size):
            if not any(subnet.overlaps(used) for used in used_subnets):
                return str(subnet)

        raise AddressManagerException('No available CIDR block found in AWS')

    def _create_vpc(self) -> str:
        # Preferable IP range for VPC is 10.0.0.0/8 according to
        # https://docs.aws.amazon.com/vpc/latest/userguide/vpc-cidr-blocks.html.
        # Do not try 172.* and 192.* as 10.* should be enough.
        network: IPv4Network | IPv6Network = ipaddress.ip_network('10.0.0.0/8')
        vpc_network_size: int = 24  # Network size 24 (255 addresses) is enough for shield
        cidr: str = self._find_available_subnet(network, self._get_vpc_networks(), vpc_network_size)
        response: CreateVpcResultTypeDef = self.ec2_client.create_vpc(
            CidrBlock=cidr,
            TagSpecifications=[{'ResourceType': 'vpc', 'Tags': [{'Key': 'Name', 'Value': 'DDosShield'}]}],
        )
        vpc_id: str = response['Vpc']['VpcId']
        self.event_processor.event('Created AWS VPC {id} with cidr={cidr}', id=vpc_id, cidr=cidr)
        try:
            self.state_manager.add_address_manager_created_object(AwsObjectTypes.VPC.value, vpc_id)
        except Exception as e:
            self._remove_vpc(vpc_id)
            raise e
        return vpc_id

    def _remove_vpc(self, vpc_id: str) -> bool:
        self.ec2_client.delete_vpc(VpcId=vpc_id)
        self.event_processor.event('Deleted AWS VPC {id}', id=vpc_id)
        self.state_manager.del_address_manager_created_object(AwsObjectTypes.VPC.value, vpc_id)
        return True

    def _create_subnet(self, vpc_id: str, cidr_block: str, availability_zone: str) -> AwsSubnetData:
        response: CreateSubnetResultTypeDef = self.ec2_client.create_subnet(
            VpcId=vpc_id, CidrBlock=cidr_block, AvailabilityZone=availability_zone
        )
        subnet = AwsSubnetData(
            subnet_id=response['Subnet']['SubnetId'], availability_zone=availability_zone, cidr_block=cidr_block
        )
        self.event_processor.event(
            'Created AWS subnet {id} with cidr={cidr} in {az} availability zone',
            id=subnet.subnet_id,
            cidr=subnet.cidr_block,
            az=subnet.availability_zone,
        )
        try:
            self.state_manager.add_address_manager_created_object(AwsObjectTypes.SUBNET.value, subnet.subnet_id)
        except Exception as e:
            self._remove_subnet(subnet.subnet_id)
            raise e
        return subnet

    def _remove_subnet(self, subnet_id: str) -> bool:
        self.ec2_client.delete_subnet(SubnetId=subnet_id)
        self.event_processor.event('Deleted AWS subnet {id}', id=subnet_id)
        self.state_manager.del_address_manager_created_object(AwsObjectTypes.SUBNET.value, subnet_id)
        return True

    def _create_target_group(self, vpc_data: AwsVpcData, server_data: AwsShieldedServerData) -> str:
        group_name: str = f'miner-target-group-{self._generate_random_alnum_string(8)}'
        # Health check can't be disabled - as for now we use traffic-port
        response: CreateTargetGroupOutputTypeDef = self.elb_client.create_target_group(
            Name=group_name,
            Protocol='HTTP',
            Port=server_data.server_location.port,
            VpcId=vpc_data.vpc_id,
            TargetType='instance',
            HealthCheckEnabled=True,
            HealthCheckProtocol='HTTP',
            HealthCheckPort='traffic-port',
        )
        target_group_id: str = response['TargetGroups'][0]['TargetGroupArn']
        self.event_processor.event('Created AWS TargetGroup, name={name}, id={id}', name=group_name, id=target_group_id)

        try:
            # TODO: AWS load balancers prohibit redirection to IP address from outside of AWS. This have to be done
            # other way - probably by creating EC2 instance and configuring there haproxy, which will redirect traffic
            # to the shielded server.
            self.elb_client.register_targets(
                TargetGroupArn=target_group_id,
                Targets=[{'Id': server_data.server_location.location_value, 'Port': server_data.server_location.port}],
            )
            self.state_manager.add_address_manager_created_object(AwsObjectTypes.TARGET_GROUP.value, target_group_id)
        except Exception as e:
            self._remove_target_group(target_group_id)
            raise e

        return target_group_id

    def _remove_target_group(self, target_group_id: str) -> bool:
        current_server_data: AwsShieldedServerData | None = self._load_server_data()
        if current_server_data:
            self.elb_client.deregister_targets(
                TargetGroupArn=target_group_id, Targets=[{'Id': current_server_data.server_location.location_value}]
            )

        error_code: str = ''
        for _ in range(self.AWS_OPERATION_MAX_RETRIES):
            try:
                self.elb_client.delete_target_group(TargetGroupArn=target_group_id)
                break
            except ClientError as e:
                error_code = e.response['Error']['Code']
                if error_code == 'ResourceInUse':
                    time.sleep(self.AWS_OPERATION_RETRY_DELAY_SEC)  # Wait for target group to be deregistered
                else:
                    raise e
        else:
            # It happens quite often and sometimes AWS waits for many minutes before allowing to remove target group.
            # But we don't want to wait for too long.
            # If it happens during tests, user should remove target group later manually using AWS panel to not leave
            # unneeded objects in AWS.
            self.event_processor.event(
                'Failed to remove AWS TargetGroup {id}, error={error_code}', id=target_group_id, error_code=error_code
            )
            return False

        self.event_processor.event('Deleted AWS TargetGroup {id}', id=target_group_id)
        self.state_manager.del_address_manager_created_object(AwsObjectTypes.TARGET_GROUP.value, target_group_id)
        return True

    def _create_elb(self, shield_subnets: list[AwsSubnetData], target_group_id: str, security_group_id: str) -> str:
        elb_name: str = f'miner-elb-{self._generate_random_alnum_string(8)}'
        subnets_ids: list[str] = [subnet.subnet_id for subnet in shield_subnets]
        response: CreateLoadBalancerOutputTypeDef = self.elb_client.create_load_balancer(
            Name=elb_name,
            Subnets=subnets_ids,
            SecurityGroups=[security_group_id],
            Scheme='internet-facing',
            Type='application',
        )
        elb_info: LoadBalancerTypeDef = response['LoadBalancers'][0]
        elb_id: str = elb_info['LoadBalancerArn']
        self.event_processor.event('Created AWS ELB, name={name}, id={id}', name=elb_name, id=elb_id)

        try:
            self.elb_client.create_listener(
                LoadBalancerArn=elb_id,
                Protocol='HTTP',
                Port=80,
                DefaultActions=[{'Type': 'forward', 'TargetGroupArn': target_group_id}],
            )
            self.state_manager.add_address_manager_created_object(AwsObjectTypes.ELB.value, elb_id)
        except Exception as e:
            self._remove_elb(elb_id)
            raise e

        return elb_id

    def _remove_elb(self, elb_id: str) -> bool:
        self.elb_client.delete_load_balancer(LoadBalancerArn=elb_id)
        self.event_processor.event('Deleted AWS ELB {id}', id=elb_id)
        self.state_manager.del_address_manager_created_object(AwsObjectTypes.ELB.value, elb_id)
        return True

    def _get_elb_info(self, elb_id: str) -> AwsELBData:
        response: DescribeLoadBalancersOutputTypeDef = self.elb_client.describe_load_balancers(
            LoadBalancerArns=[elb_id]
        )
        elb_info: LoadBalancerTypeDef = response['LoadBalancers'][0]
        return AwsELBData(
            id=elb_info['LoadBalancerArn'],
            dns_name=elb_info['DNSName'],
            hosted_zone_id=elb_info['CanonicalHostedZoneId'],
        )

    def _create_security_group(self, vpc_data: AwsVpcData, server_port: int) -> str:
        group_name: str = f'miner-security-group-{self._generate_random_alnum_string(8)}'
        response: CreateSecurityGroupResultTypeDef = self.ec2_client.create_security_group(
            GroupName=group_name, Description='Security group for miner instance', VpcId=vpc_data.vpc_id
        )
        security_group_id: str = response['GroupId']
        self.event_processor.event(
            'Created AWS SecurityGroup, name={name}, id={id}', name=group_name, id=security_group_id
        )

        try:
            self.ec2_client.authorize_security_group_ingress(
                GroupId=security_group_id,
                IpPermissions=[
                    {'FromPort': 80, 'ToPort': server_port, 'IpProtocol': 'tcp', 'IpRanges': [{'CidrIp': '0.0.0.0/0'}]}
                ],
            )
            self.state_manager.add_address_manager_created_object(
                AwsObjectTypes.SECURITY_GROUP.value, security_group_id
            )
        except Exception as e:
            self._remove_security_group(security_group_id)
            raise e

        return security_group_id

    def _remove_security_group(self, security_group_id: str) -> bool:
        error_code: str = ''
        for _ in range(self.AWS_OPERATION_MAX_RETRIES):
            try:
                self.ec2_client.delete_security_group(GroupId=security_group_id)
                break
            except ClientError as e:
                error_code = e.response['Error']['Code']
                if error_code == 'DependencyViolation':
                    time.sleep(self.AWS_OPERATION_RETRY_DELAY_SEC)  # Wait for ELB to be removed
                else:
                    raise e
        else:
            # It happens quite often and sometimes AWS waits for many minutes before allowing to remove security group.
            # But we don't want to wait for too long.
            # If it happens during tests, user should remove security group later manually using AWS panel to not leave
            # unneeded objects in AWS.
            self.event_processor.event(
                'Failed to remove AWS SecurityGroup {id}, error={error_code}',
                id=security_group_id,
                error_code=error_code,
            )
            return False

        self.event_processor.event('Deleted AWS SecurityGroup {id}', id=security_group_id)
        self.state_manager.del_address_manager_created_object(AwsObjectTypes.SECURITY_GROUP.value, security_group_id)
        return True

    def _delete_route53_records(self, hosted_zone_id: str):
        address_manager_created_objects: MappingProxyType[str, frozenset[str]] = (
            self.state_manager.get_state().address_manager_created_objects
        )
        if AwsObjectTypes.DNS_ENTRY.value not in address_manager_created_objects:
            return

        created_entries: frozenset[str] = self.state_manager.get_state().address_manager_created_objects[
            AwsObjectTypes.DNS_ENTRY.value
        ]
        hosted_zone = self.route53_client.get_hosted_zone_by_id(hosted_zone_id)
        for record_set in hosted_zone.record_sets:
            if codecs.decode(record_set.name, 'unicode_escape') in created_entries:
                self._delete_route53_record(record_set, hosted_zone)

        # Clean from state entries without working address (without corresponding record_set in hosted_zone)
        address_manager_created_objects = self.state_manager.get_state().address_manager_created_objects
        if AwsObjectTypes.DNS_ENTRY.value not in address_manager_created_objects:
            return
        for created_entry in address_manager_created_objects[AwsObjectTypes.DNS_ENTRY.value]:
            self.state_manager.del_address_manager_created_object(AwsObjectTypes.DNS_ENTRY.value, created_entry)

    def _create_firewall(self) -> str:
        waf_name: str = f'miner-waf-{self._generate_random_alnum_string(8)}'
        response: CreateWebACLResponseTypeDef = self.waf_client.create_web_acl(
            Name=waf_name,
            Scope='REGIONAL',
            DefaultAction={'Block': {}},
            Rules=[],
            VisibilityConfig={'SampledRequestsEnabled': True, 'CloudWatchMetricsEnabled': True, 'MetricName': waf_name},
        )
        waf_arn: str = response['Summary']['ARN']
        self.event_processor.event('Created AWS WAF, name={name}, id={id}', name=waf_name, id=waf_arn)

        assert self.elb_data is not None, '_validate_manager_state creates ELB and should be called before'

        error_code: str = ''
        for _ in range(self.AWS_OPERATION_MAX_RETRIES):
            try:
                self.waf_client.associate_web_acl(WebACLArn=waf_arn, ResourceArn=self.elb_data.id)
                self.event_processor.event(
                    'Associated AWS WAF {waf_id} to ELB {elb_id}', waf_id=waf_arn, elb_id=self.elb_data.id
                )
                break
            except ClientError as e:
                error_code = e.response['Error']['Code']
                time.sleep(self.AWS_OPERATION_RETRY_DELAY_SEC)  # Wait for WAF to be created
        else:
            self._remove_firewall(waf_arn)
            # It happens quite often and sometimes creation of ELB propagates for many minutes before association is
            # allowed. But we don't want to wait for too long.
            raise AddressManagerException(
                f'Failed to associate AWS WAF {waf_arn} with ELB {self.elb_data.id}, error={error_code}'
            )

        try:
            self.state_manager.add_address_manager_created_object(AwsObjectTypes.WAF.value, waf_arn)
        except Exception as e:
            self._remove_firewall(waf_arn)
            raise e

        return waf_arn

    def _remove_firewall(self, waf_arn: str) -> bool:
        created_objects: MappingProxyType[str, frozenset[str]] = (
            self.state_manager.get_state().address_manager_created_objects
        )

        if AwsObjectTypes.ELB.value in created_objects:
            assert len(created_objects[AwsObjectTypes.ELB.value]) == 1, 'only one ELB should be created'
            elb_id: str = next(iter(created_objects[AwsObjectTypes.ELB.value]))
            self.waf_client.disassociate_web_acl(ResourceArn=elb_id)

        waf_data: GetWebACLResponseTypeDef = self._get_firewall_info(waf_arn)
        acl_data: WebACLTypeDef = waf_data['WebACL']
        waf_id: str = acl_data['Id']
        lock_token = waf_data['LockToken']

        error_code: str = ''
        for _ in range(self.AWS_OPERATION_MAX_RETRIES):
            try:
                self.waf_client.delete_web_acl(Name=acl_data['Name'], Id=waf_id, Scope='REGIONAL', LockToken=lock_token)
                break
            except ClientError as e:
                error_code = e.response['Error']['Code']
                if error_code == 'WAFAssociatedItemException':
                    time.sleep(self.AWS_OPERATION_RETRY_DELAY_SEC)  # Wait for ELB disassociating
                else:
                    raise e
        else:
            # It happens quite often and sometimes AWS waits for many minutes before allowing to remove WAF.
            # But we don't want to wait for too long.
            # If it happens during tests, user should remove WAF later manually using AWS panel to not leave
            # unneeded objects in AWS.
            self.event_processor.event(
                'Failed to remove AWS WAF {id}, error={error_code}', id=waf_id, error_code=error_code
            )
            return False

        self.event_processor.event('Deleted AWS WAF {id}', id=waf_arn)
        self.state_manager.del_address_manager_created_object(AwsObjectTypes.WAF.value, waf_arn)
        return True

    def _get_firewall_info(self, waf_arn: str) -> GetWebACLResponseTypeDef:
        waf_name: str = self._get_name_from_waf_arn(waf_arn)
        waf_id: str = self._get_id_from_waf_arn(waf_arn)
        return self.waf_client.get_web_acl(Name=waf_name, Id=waf_id, Scope='REGIONAL')

    def _update_web_acl(self, waf_data: GetWebACLResponseTypeDef, rules: Sequence[RuleTypeDef]):
        acl_data: WebACLTypeDef = waf_data['WebACL']
        lock_token = waf_data['LockToken']
        self.waf_client.update_web_acl(
            Name=acl_data['Name'],
            Id=acl_data['Id'],
            Scope='REGIONAL',
            DefaultAction=acl_data['DefaultAction'],  # type: ignore
            Rules=rules,
            VisibilityConfig=acl_data['VisibilityConfig'],
            LockToken=lock_token,
        )

    def _add_domain_rule_to_firewall(self, waf_arn: str, domain: str):
        waf_data: GetWebACLResponseTypeDef = self._get_firewall_info(waf_arn)
        rules: list[RuleTypeDef] = waf_data['WebACL']['Rules']  # type: ignore
        rule_name: str = f'miner-waf-rule-{self._generate_random_alnum_string(8)}'
        priority: int = rules[-1]['Priority'] + 1 if rules else 1
        rule: RuleTypeDef = {
            'Name': rule_name,
            'Priority': priority,
            'Statement': {
                'ByteMatchStatement': {
                    'SearchString': domain,
                    'FieldToMatch': {'SingleHeader': {'Name': 'host'}},
                    'TextTransformations': [{'Priority': 0, 'Type': 'NONE'}],
                    'PositionalConstraint': 'EXACTLY',
                }
            },
            'Action': {'Allow': {}},
            'VisibilityConfig': {
                'SampledRequestsEnabled': True,
                'CloudWatchMetricsEnabled': True,
                'MetricName': rule_name,
            },
        }
        rules.append(rule)
        self._update_web_acl(waf_data, rules)
        self.event_processor.event(
            'Added rule {rule_name} to AWS WAF {waf_id}, domain={domain}',
            rule_name=rule_name,
            waf_id=waf_arn,
            domain=domain,
        )

    def _remove_domain_rule_from_firewall(self, waf_arn: str, domain: str):
        waf_data: GetWebACLResponseTypeDef = self._get_firewall_info(waf_arn)
        rules: list[RuleOutputTypeDef] = waf_data['WebACL']['Rules']
        rule: RuleOutputTypeDef | None = self._find_rule(rules, domain)
        if rule is None:
            return
        rules.remove(rule)
        self._update_web_acl(waf_data, rules)  # type: ignore
        self.event_processor.event(
            'Removed rule {rule_name} from AWS WAF {waf_id}, domain={domain}',
            rule_name=rule['Name'],
            waf_id=waf_arn,
            domain=domain,
        )

    def _clear_firewall_rules(self, waf_arn: str):
        waf_data: GetWebACLResponseTypeDef = self._get_firewall_info(waf_arn)
        self._update_web_acl(waf_data, [])

    def _init_hosted_zone(self):
        # Add wildcard subdomain to hosted zone - Host header validation is done by WAF
        self._add_route53_record('*', self.hosted_zone)

    @classmethod
    def _find_rule(cls, rules: list[RuleOutputTypeDef], domain: str) -> RuleOutputTypeDef | None:
        for rule in rules:
            try:
                rule_domain: str = rule['Statement']['ByteMatchStatement']['SearchString'].decode()
            except KeyError:
                continue
            if rule_domain == domain:
                return rule
        return None

    @classmethod
    def _get_id_from_waf_arn(cls, waf_arn: str) -> str:
        return waf_arn.split('/')[-1]

    @classmethod
    def _get_name_from_waf_arn(cls, waf_arn: str) -> str:
        return waf_arn.split('/')[-2]

    def _create_subnets_if_needed(
        self,
        vpc_data: AwsVpcData,
        server_data: AwsShieldedServerData,
        created_objects: MappingProxyType[str, frozenset[str]],
    ) -> list[AwsSubnetData]:
        shield_subnets: list[AwsSubnetData] = []
        if AwsObjectTypes.SUBNET.value in created_objects:
            for subnet_id in created_objects[AwsObjectTypes.SUBNET.value]:
                shield_subnets.append(self._get_subnet_data(subnet_id))
        if server_data.aws_location:
            shield_subnets.append(self._get_subnet_data(server_data.aws_location.subnet_id))

        # ELB needs at least two subnets in different AZs
        min_subnets_for_elb: int = 2
        if len(shield_subnets) >= min_subnets_for_elb:
            return shield_subnets

        response: DescribeAvailabilityZonesResultTypeDef = self.ec2_client.describe_availability_zones()
        vpc_availability_zones: set[str] = {az['ZoneName'] for az in response['AvailabilityZones']}
        subnets_availability_zones: set[str] = {subnet.availability_zone for subnet in shield_subnets}
        # Region should have at least 2 different availability zones, and we need to create at most 2 subnets, so
        # we can use set difference to find remaining availability zones.
        remaining_availability_zones: set[str] = vpc_availability_zones - subnets_availability_zones

        vpc_network: IPv4Network | IPv6Network = ipaddress.ip_network(vpc_data.cidr_block)
        used_networks: list[IPv4Network | IPv6Network] = self._get_subnet_networks(vpc_data)
        # 27 as specified in https://docs.aws.amazon.com/elasticloadbalancing/latest/application/application-load-balancers.html#availability-zones  # noqa: E501
        min_elb_subnet_size: int = 27

        while len(shield_subnets) < min_subnets_for_elb:
            new_subnet_cidr: str = self._find_available_subnet(vpc_network, used_networks, min_elb_subnet_size)
            new_subnet_az: str = remaining_availability_zones.pop()
            new_subnet: AwsSubnetData = self._create_subnet(vpc_data.vpc_id, new_subnet_cidr, new_subnet_az)
            used_networks.append(ipaddress.ip_network(new_subnet_cidr))
            shield_subnets.append(new_subnet)
        return shield_subnets

    def _create_target_group_if_needed(
        self,
        vpc_data: AwsVpcData,
        server_data: AwsShieldedServerData,
        created_objects: MappingProxyType[str, frozenset[str]],
    ) -> str:
        if AwsObjectTypes.TARGET_GROUP.value in created_objects:
            assert len(created_objects[AwsObjectTypes.TARGET_GROUP.value]) == 1, 'only one group should be created'
            target_group_id: str = next(iter(created_objects[AwsObjectTypes.TARGET_GROUP.value]))
            return target_group_id

        return self._create_target_group(vpc_data, server_data)

    def _create_security_group_if_needed(
        self, vpc_data: AwsVpcData, server_port: int, created_objects: MappingProxyType[str, frozenset[str]]
    ) -> str:
        if AwsObjectTypes.SECURITY_GROUP.value in created_objects:
            assert len(created_objects[AwsObjectTypes.SECURITY_GROUP.value]) == 1, 'only one group should be created'
            security_group_id: str = next(iter(created_objects[AwsObjectTypes.SECURITY_GROUP.value]))
            return security_group_id

        return self._create_security_group(vpc_data, server_port)

    def _create_vpc_if_needed(
        self, server_data: AwsShieldedServerData, created_objects: MappingProxyType[str, frozenset[str]]
    ) -> AwsVpcData:
        vpc_id: str
        if server_data.aws_location:
            assert AwsObjectTypes.VPC.value not in created_objects, 'VPC of EC2 instance should be used'
            vpc_id = server_data.aws_location.vpc_id
        else:
            if AwsObjectTypes.VPC.value in created_objects:
                assert len(created_objects[AwsObjectTypes.VPC.value]) == 1, 'only one VPC should be created'
                vpc_id = next(iter(created_objects[AwsObjectTypes.VPC.value]))
            else:
                vpc_id = self._create_vpc()

        return self._get_vpc_data(vpc_id)

    def _create_elb_if_needed(self, server_data: AwsShieldedServerData) -> AwsELBData:
        created_objects: MappingProxyType[str, frozenset[str]] = (
            self.state_manager.get_state().address_manager_created_objects
        )

        if AwsObjectTypes.ELB.value in created_objects:
            assert len(created_objects[AwsObjectTypes.ELB.value]) == 1, 'only one ELB should be created'
            return self._get_elb_info(next(iter(created_objects[AwsObjectTypes.ELB.value])))

        server_port: int = server_data.server_location.port
        vpc_data: AwsVpcData = self._create_vpc_if_needed(server_data, created_objects)
        shield_subnets: list[AwsSubnetData] = self._create_subnets_if_needed(vpc_data, server_data, created_objects)
        target_group_id: str = self._create_target_group_if_needed(vpc_data, server_data, created_objects)
        security_group_id: str = self._create_security_group_if_needed(vpc_data, server_port, created_objects)
        elb_id: str = self._create_elb(shield_subnets, target_group_id, security_group_id)
        return self._get_elb_info(elb_id)

    def _create_firewall_if_needed(self) -> str:
        created_objects: MappingProxyType[str, frozenset[str]] = (
            self.state_manager.get_state().address_manager_created_objects
        )
        if AwsObjectTypes.WAF.value in created_objects:
            assert len(created_objects[AwsObjectTypes.WAF.value]) == 1, 'only one firewall should be created'
            return next(iter(created_objects[AwsObjectTypes.WAF.value]))

        return self._create_firewall()

    def _init_hosted_zone_if_needed(self) -> None:
        created_objects: MappingProxyType[str, frozenset[str]] = (
            self.state_manager.get_state().address_manager_created_objects
        )
        if AwsObjectTypes.DNS_ENTRY.value in created_objects:
            assert len(created_objects[AwsObjectTypes.DNS_ENTRY.value]) == 1, 'only one entry should be created'
            return

        self._init_hosted_zone()



from __future__ import annotations

import argparse
import asyncio
import logging
import re
import sys
import threading
from abc import ABC, abstractmethod
from queue import Queue
from time import sleep
from typing import TYPE_CHECKING

from pydantic import BaseModel, Field
from pydantic_settings import BaseSettings

from bt_ddos_shield.address_manager import (
    AbstractAddressManager,
    AwsAddressManager,
    ShieldedServerLocation,
    ShieldedServerLocationType,
)
from bt_ddos_shield.blockchain_manager import (
    AbstractBlockchainManager,
    BittensorBlockchainManager,
)
from bt_ddos_shield.encryption_manager import AbstractEncryptionManager, ECIESEncryptionManager
from bt_ddos_shield.event_processor import AbstractMinerShieldEventProcessor, PrintingMinerShieldEventProcessor
from bt_ddos_shield.manifest_manager import (
    AbstractManifestManager,
    AbstractManifestSerializer,
    JsonManifestSerializer,
    Manifest,
    ManifestDeserializationException,
    ManifestNotFoundException,
    S3ManifestManager,
)
from bt_ddos_shield.state_manager import (
    AbstractMinerShieldStateManager,
    MinerShieldState,
    SQLAlchemyMinerShieldStateManager,
)
from bt_ddos_shield.utils import AWSClientFactory, Hotkey, PublicKey, ShieldAddress, SubtensorSettings, WalletSettings
from bt_ddos_shield.validators_manager import AbstractValidatorsManager, BittensorValidatorsManager

if TYPE_CHECKING:
    from collections.abc import Iterable
    from types import MappingProxyType


class MinerShieldOptions(BaseModel):
    """
    A class to represent the configuration options for the MinerShield.
    """

    auto_hide_original_server: bool = False
    """
    If True, the original server will be hidden after some time after shield gets enabled. Method hide_original_server
    in AddressManager will be called.
    """

    auto_hide_delay_sec: int = 1200
    """ Time in seconds after which the original server will be hidden if auto_hide_original_server is set to True. """

    retry_limit: int | None = None
    """
    Limit of retries for failed tasks. If None, task will be retried forever. Should be changed only for testing.
    Retrying can be also stopped by disabling shield (with Ctrl-C or calling disable).
    """

    retry_delay_sec: int = 5
    """ Time in seconds to wait before retrying failed task. """

    validate_interval_sec: int = 120
    """
    Time in seconds between cyclic calls of validate task. It checks if every part of shield is working
    correctly.
    """


class MinerShieldException(Exception):
    pass


class MinerShieldDisabledException(MinerShieldException):
    """
    Exception raised when shield is disabled and user want to schedule some action to it.
    """


class MinerShield:
    """
    Main class to be used by Miner to shield himself from DDoS. Call enable() to start the shield. No methods in
    managers should be called directly. All operations are done by worker thread. After starting shield user can
    schedule tasks to be executed asynchronously.
    """

    worker_thread: threading.Thread | None  # main thread executing tasks
    task_queue: Queue[AbstractMinerShieldTask]  # queue of tasks to be executed
    run: bool  # flag meaning if shield is running
    finishing: bool  # flag meaning if shield should finish its work
    ticker: threading.Event  # ticker to be used in ticker_thread
    ticker_thread: threading.Thread | None  # thread used to schedule validate operation cyclically
    validators_manager: AbstractValidatorsManager  # used to manage validators and their keys
    address_manager: AbstractAddressManager  # used to manage public IP/domain addresses assigned to validators
    manifest_manager: AbstractManifestManager  # used to manage publishing manifest file
    blockchain_manager: AbstractBlockchainManager  # used to manage blockchain operations
    state_manager: AbstractMinerShieldStateManager  # used to manage state of the shield
    event_processor: AbstractMinerShieldEventProcessor  # used to handle events generated by the shield
    options: MinerShieldOptions  # configuration options for shield

    def __init__(
        self,
        validators_manager: AbstractValidatorsManager,
        address_manager: AbstractAddressManager,
        manifest_manager: AbstractManifestManager,
        blockchain_manager: AbstractBlockchainManager,
        state_manager: AbstractMinerShieldStateManager,
        event_processor: AbstractMinerShieldEventProcessor,
        options: MinerShieldOptions,
    ):
        self.validators_manager = validators_manager
        self.address_manager = address_manager
        self.manifest_manager = manifest_manager
        self.blockchain_manager = blockchain_manager
        self.state_manager = state_manager
        self.event_processor = event_processor
        self.options = options

        self.worker_thread = None
        self.task_queue = Queue()
        self.run = False
        self.finishing = False
        self.ticker = threading.Event()
        self.ticker_thread = None

    def enable(self):
        """
        Enable shield. It starts worker thread, which will do such steps if run for the first time:
        1. Fetch validators keys.
        2. Creates addresses for all validators.
        3. Save manifest file.
        4. Publish link to manifest file to blockchain.
        5. Eventually close public access to original IP after some time.

        It puts events to event_manager after each finished operation. Current state is managed by state_manager.
        If any error occurs it is retried forever until shield is disabled.

        When shield is running, user can schedule tasks to be processed by worker.
        """
        if self.worker_thread is not None:
            # already started
            return

        self.finishing = False
        self.ticker.clear()
        self.run = True
        self._add_task(MinerShieldInitializeTask())
        self.worker_thread = threading.Thread(target=self._worker_function)
        self.worker_thread.start()

    def disable(self):
        """
        Disable shield. It stops worker thread after finishing current task. Function blocks until worker is stopped.
        """
        self._add_task(MinerShieldDisableTask())

        self.finishing = True
        self.ticker.set()

        if self.worker_thread is not None:
            self.worker_thread.join()
            self.worker_thread = None

        # Clear tasks possibly added after MinerShieldDisableTask and before setting finishing to True to allow
        # ticker_thread to finish.
        self._clear_tasks()

        if self.ticker_thread is not None:
            self.ticker_thread.join()
            self.ticker_thread = None

        # As we don't use any mutex, there is possibility that ticker_thread adds validation task after above
        # _clear_tasks call. This can happen when ticker_thread execution context was switched inside _add_task method
        # after checking self.finishing flag, but before really putting task to queue and whole disable method run
        # (until self.ticker_thread.join()) before switching context back to ticker_thread.
        self._clear_tasks()

    def ban_validator(self, validator_hotkey: Hotkey):
        """
        Ban a validator by its hotkey. Task will be executed by worker. It will update manifest file and publish info
        about new file version to blockchain.
        """
        self._add_task(MinerShieldBanValidatorTask(validator_hotkey))

    def _add_task(self, task: AbstractMinerShieldTask):
        """
        Add task to task queue. It will be handled by _worker_function.
        """
        if not self.run or self.finishing:
            raise MinerShieldDisabledException()

        self.task_queue.put(task)

    def _clear_tasks(self):
        while not self.task_queue.empty():
            self.task_queue.get()
            self.task_queue.task_done()

    def _worker_function(self) -> None:
        """
        Function called in separate thread by enable() to start the shield. It is handling events put to task_queue.
        """
        self._event('Starting shield')

        while self.run:
            task: AbstractMinerShieldTask = self.task_queue.get()
            try_count: int = 1

            while self.run:
                self._event('Handling task {task}, try {try_count}', task=task, try_count=try_count)

                try:
                    task.run(self)
                    self._event('Task {task} finished successfully', task=task)
                    break
                except Exception as e:
                    self._event('Error during handling task {task}', e, task=task)

                    if self.finishing:
                        break

                    if self.options.retry_limit and try_count > self.options.retry_limit:
                        self._event('Retry limit reached for task {task}', task=task)
                        break

                    try_count += 1
                    sleep(self.options.retry_delay_sec)

            self.task_queue.task_done()

        self._event('Stopping shield')

    def _ticker_function(self):
        while not self.ticker.wait(self.options.validate_interval_sec):
            # Wait for finishing previous tasks before adding new validation task. This is mostly to prevent
            # having multiple validation tasks in the queue.
            self.task_queue.join()
            self._add_task(MinerShieldValidateStateTask())

    def _handle_initialize(self):
        self._handle_validate_state(first_run=True)
        self.ticker_thread = threading.Thread(target=self._ticker_function)
        self.ticker_thread.start()

    def _reload_state(self, first_run: bool):
        try:
            self.state_manager.get_state(reload=True)
            self._event('State reloaded')
        except Exception as e:
            self._event('Error during reloading state', e)
            if first_run:
                # We cannot continue without initial shield state
                raise e

    def _reload_validators(self, first_run: bool) -> bool:
        """
        Reload validators from validators manager.

        Returns:
             bool: If validators set is different from one currently stored in shield state.
        """
        try:
            self.validators_manager.reload_validators()
            fetched_validators: MappingProxyType[Hotkey, PublicKey] = self.validators_manager.get_validators()

            deprecated_validators, new_validators, changed_validators = self._calculate_validators_diff(
                self.state_manager.get_state(), fetched_validators
            )
            validators_changed: bool = bool(deprecated_validators) or bool(new_validators) or bool(changed_validators)

            self._event(
                'Validators reloaded, got {validators_count} validators, validators_changed={validators_changed}',
                validators_count=len(fetched_validators),
                validators_changed=validators_changed,
            )

            return validators_changed
        except Exception as e:
            self._event('Error during reloading validators', e)
            if first_run:
                # We cannot continue without initial validators state
                raise e

        return False

    def _validate_addresses(self, first_run: bool) -> bool:
        """
        Validate addresses used by validators. Remove validators with invalid addresses from shield state - they will
        be added again with new addresses in a moment.

        Returns:
             bool: If any validators were removed.
        """
        validators_changed: bool = False
        try:
            current_state: MinerShieldState = self.state_manager.get_state()
            invalid_addresses: set[Hotkey] = self.address_manager.validate_addresses(current_state.validators_addresses)
            if invalid_addresses:
                self._event(
                    'Removing invalid addresses for given validators: {invalid_addresses}',
                    invalid_addresses=invalid_addresses,
                )
                for validator in invalid_addresses:
                    self.state_manager.remove_validator(validator)
                    validators_changed = True
        except Exception as e:
            self._event('Error during validating addresses', e)
            if first_run:
                # We cannot continue without initializing address_manager
                raise e
            # If error happens later, just proceed. It is only validation, and it will be called again
            # by _ticker_function.

        return validators_changed

    def _validate_manifest_file(self) -> bool:
        """
        Validate manifest file by comparing uploaded content with expected one. Returns if it is valid.
        """
        try:
            current_state: MinerShieldState = self.state_manager.get_state()
            current_manifest: Manifest = asyncio.run(
                self.manifest_manager.get_manifest(self.manifest_manager.get_manifest_url())
            )
            new_manifest: Manifest = self.manifest_manager.create_manifest(
                current_state.validators_addresses, current_state.known_validators
            )
            same_content: bool = new_manifest.md5_hash == current_manifest.md5_hash
            self._event('Manifest file validation finished, same content={same_content}', same_content=same_content)
            return same_content
        except (ManifestNotFoundException, ManifestDeserializationException):
            return False
        except Exception as e:
            self._event('Error during validating manifest file', e)
            # If error happened, assume that manifest file is valid - it is only validation, and it will be called
            # again by _ticker_function
            return True

    def _handle_validate_state(self, first_run: bool = False):
        """
        Refresh all data and validate current state of shield. If shield is not yet fully working, it will finish
        startup process. If something is wrong, it will be fixed.
        """
        self._reload_state(first_run)
        validators_changed: bool = self._reload_validators(first_run)
        if self._validate_addresses(first_run):
            validators_changed = True

        if validators_changed:
            self._add_task(MinerShieldValidatorsChangedTask())
            # do not validate manifest, because it will be updated in the moment
            return

        if not self._validate_manifest_file():
            self._add_task(MinerShieldUpdateManifestTask())
            # do not validate manifest info, because it will be updated in the moment
            return

        # this will validate manifest info and will publish new one only if needed
        self._add_task(MinerShieldPublishManifestTask())

    def _handle_disable(self):
        self.run = False

    @classmethod
    def _calculate_validators_diff(
        cls, current_state: MinerShieldState, fetched_validators: MappingProxyType[Hotkey, PublicKey]
    ):
        """
        Calculates difference between newly fetched validators set and one saved in state and returns validators
        which should be removed, added or updated.
        """
        # remove banned validators from fetched validators
        fetched_not_banned_validators: dict[Hotkey, PublicKey] = dict(fetched_validators)
        for banned_validator in current_state.banned_validators.keys():
            fetched_not_banned_validators.pop(banned_validator, None)

        # calculate difference between current state and fetched validators
        deprecated_validators = current_state.known_validators.keys() - fetched_not_banned_validators.keys()
        new_validators = fetched_not_banned_validators.keys() - current_state.known_validators.keys()
        common_validators = fetched_not_banned_validators.keys() & current_state.known_validators.keys()
        changed_validators = {
            k: fetched_not_banned_validators[k]
            for k in common_validators
            if fetched_not_banned_validators[k] != current_state.known_validators[k]
        }

        return deprecated_validators, new_validators, changed_validators

    def _handle_deprecated_validators(self, current_state: MinerShieldState, deprecated_validators: set[Hotkey]):
        for validator in deprecated_validators:
            self._event('Removing validator {validator}', validator=validator)

            active_validator_addresses = current_state.validators_addresses
            if validator in active_validator_addresses:
                self.address_manager.remove_address(active_validator_addresses[validator])

            self.state_manager.remove_validator(validator)

    def _handle_new_validators(
        self, fetched_validators: MappingProxyType[Hotkey, PublicKey], new_validators: set[Hotkey]
    ):
        for validator in new_validators:
            self._event('Adding validator {validator}', validator=validator)

            new_address: ShieldAddress = self.address_manager.create_address(validator)

            try:
                self.state_manager.add_validator(validator, fetched_validators[validator], new_address)
            except Exception as e:
                self.address_manager.remove_address(new_address)
                raise e

    def _handle_changed_validators(self, changed_validators: dict[Hotkey, PublicKey]):
        for validator, new_key in changed_validators.items():
            self._event('Updating validator {validator}', validator=validator)
            self.state_manager.update_validator_public_key(validator, new_key)

    def _handle_validators_change(self) -> None:
        current_state: MinerShieldState = self.state_manager.get_state()
        fetched_validators: MappingProxyType[Hotkey, PublicKey] = self.validators_manager.get_validators()

        deprecated_validators, new_validators, changed_validators = self._calculate_validators_diff(
            current_state, fetched_validators
        )

        self._event(
            'Handling validators change, deprecated_validators count={deprecated_validators_count}'
            ', new_validators count={new_validators_count}, changed_validators count={changed_validators_count}',
            deprecated_validators_count=len(deprecated_validators),
            new_validators_count=len(new_validators),
            changed_validators_count=len(changed_validators),
        )

        self._handle_deprecated_validators(current_state, deprecated_validators)
        self._handle_new_validators(fetched_validators, new_validators)
        self._handle_changed_validators(changed_validators)

        if deprecated_validators or new_validators or changed_validators:
            # if anything changed update manifest file and publish new version to blockchain
            self._add_task(MinerShieldUpdateManifestTask())

    def _handle_ban_validator(self, validator_hotkey: Hotkey):
        """
        Ban validator by its hotkey. If something changed, MinerShieldValidatorsChangedTask will apply asynchronously
        this change where needed.
        """
        self.state_manager.add_banned_validator(validator_hotkey)
        self._event('Validator {validator_hotkey} added to banned set', validator_hotkey=validator_hotkey)
        self._add_task(MinerShieldValidatorsChangedTask())

    def _handle_update_manifest(self) -> None:
        """
        Update manifest file and schedule publishing it to blockchain.
        """
        current_state: MinerShieldState = self.state_manager.get_state()
        manifest: Manifest = self.manifest_manager.create_manifest(
            current_state.validators_addresses, current_state.known_validators
        )
        self.manifest_manager.upload_manifest(manifest)
        self._event('Manifest updated, new address: {address}', address=self.manifest_manager.get_manifest_url())
        self._add_task(MinerShieldPublishManifestTask())

    def _handle_publish_manifest(self) -> None:
        """
        Publish info about current manifest file to blockchain if it is not already there.
        """
        expected_url: str = self.manifest_manager.get_manifest_url()
        current_url: str | None = asyncio.run(self.blockchain_manager.get_own_manifest_url())
        if current_url == expected_url:
            self._event('Manifest address already published')
        else:
            self.blockchain_manager.put_manifest_url(expected_url)
            self._event('Manifest published')

    def _event(self, template: str, exception: Exception | None = None, **kwargs):
        return self.event_processor.event(template, exception, **kwargs)


class AbstractMinerShieldTask(ABC):
    """
    Task to be executed by shield worker.
    """

    NAME_DELETER = re.compile(r'^MinerShield(.*)Task$')

    @abstractmethod
    def run(self, miner_shield: MinerShield):
        """
        Run task in miner_shield context.
        """
        pass

    def __repr__(self):
        return self.NAME_DELETER.sub(r'\1', self.__class__.__name__)


class MinerShieldInitializeTask(AbstractMinerShieldTask):
    def run(self, miner_shield: MinerShield):
        # noinspection PyProtectedMember
        miner_shield._handle_initialize()


class MinerShieldDisableTask(AbstractMinerShieldTask):
    def run(self, miner_shield: MinerShield):
        # noinspection PyProtectedMember
        miner_shield._handle_disable()


class MinerShieldValidateStateTask(AbstractMinerShieldTask):
    def run(self, miner_shield: MinerShield):
        # noinspection PyProtectedMember
        miner_shield._handle_validate_state()


class MinerShieldValidatorsChangedTask(AbstractMinerShieldTask):
    def run(self, miner_shield: MinerShield):
        # noinspection PyProtectedMember
        miner_shield._handle_validators_change()


class MinerShieldBanValidatorTask(AbstractMinerShieldTask):
    def __init__(self, validator_hotkey: Hotkey):
        self.validator_hotkey = validator_hotkey

    def run(self, miner_shield: MinerShield):
        # noinspection PyProtectedMember
        miner_shield._handle_ban_validator(self.validator_hotkey)


class MinerShieldUpdateManifestTask(AbstractMinerShieldTask):
    def run(self, miner_shield: MinerShield):
        # noinspection PyProtectedMember
        miner_shield._handle_update_manifest()


class MinerShieldPublishManifestTask(AbstractMinerShieldTask):
    def run(self, miner_shield: MinerShield):
        # noinspection PyProtectedMember
        miner_shield._handle_publish_manifest()


class ShieldSettings(BaseSettings):
    aws_access_key_id: str = Field(min_length=1)
    aws_secret_access_key: str = Field(min_length=1)
    aws_region_name: str = Field(min_length=1)
    """AWS region name where shield will be created"""
    aws_s3_bucket_name: str = Field(min_length=1)
    """AWS S3 bucket name where manifest file will be stored"""
    aws_route53_hosted_zone_id: str = Field(min_length=1)
    """AWS Route53 hosted zone ID for creating DNS entries for server clients"""
    aws_miner_instance_id: str = ''
    """AWS instance ID of miner server"""
    aws_miner_instance_ip: str = ''
    """AWS instance IP of miner server"""
    miner_instance_port: int
    """Port on which miner server is listening"""
    sql_alchemy_db_url: str = Field('sqlite:///ddos_shield.db', min_length=1)
    """SQL Alchemy URL to database where shield state is stored"""
    options: MinerShieldOptions = MinerShieldOptions()

    netuid: int
    subtensor: SubtensorSettings = SubtensorSettings()
    wallet: WalletSettings = WalletSettings()

    model_config = {
        'env_file': '.env',
        'env_nested_delimiter': '__',
    }


class MinerShieldFactory:
    """
    Factory class to create proper MinerShield instance basing on set environmental variables.
    """

    @classmethod
    def create_miner_shield(cls, settings: ShieldSettings, validators: Iterable[Hotkey] | None = None) -> MinerShield:
        """
        Args:
            settings: ShieldSettings instance.
            validators: Dictionary containing validators hotkeys and their public keys.
        """
        validators_manager: AbstractValidatorsManager = cls.create_validators_manager(settings, validators)
        event_processor: AbstractMinerShieldEventProcessor = cls.create_event_processor()
        state_manager: AbstractMinerShieldStateManager = cls.create_state_manager(settings)
        aws_client_factory: AWSClientFactory = cls.create_aws_client_factory(settings)
        address_manager: AbstractAddressManager = cls.create_address_manager(
            settings, aws_client_factory, event_processor, state_manager
        )
        encryption_manager: AbstractEncryptionManager = cls.create_encryption_manager()
        manifest_manager: AbstractManifestManager = cls.create_manifest_manager(
            settings, encryption_manager, aws_client_factory, event_processor
        )
        blockchain_manager: AbstractBlockchainManager = cls.create_blockchain_manager(settings, event_processor)

        if settings.options.auto_hide_original_server:
            raise MinerShieldException('Autohiding is not implemented yet')

        return MinerShield(
            validators_manager,
            address_manager,
            manifest_manager,
            blockchain_manager,
            state_manager,
            event_processor,
            settings.options,
        )

    @classmethod
    def create_validators_manager(
        cls,
        settings: ShieldSettings,
        validators: Iterable[Hotkey] | None = None,
    ) -> AbstractValidatorsManager:
        return BittensorValidatorsManager(
            subtensor=settings.subtensor.client,
            netuid=settings.netuid,
            validators=validators,
        )

    @classmethod
    def create_event_processor(cls) -> AbstractMinerShieldEventProcessor:
        return PrintingMinerShieldEventProcessor()

    @classmethod
    def create_state_manager(cls, settings: ShieldSettings) -> AbstractMinerShieldStateManager:
        return SQLAlchemyMinerShieldStateManager(settings.sql_alchemy_db_url)

    @classmethod
    def create_aws_client_factory(cls, settings: ShieldSettings) -> AWSClientFactory:
        return AWSClientFactory(settings.aws_access_key_id, settings.aws_secret_access_key, settings.aws_region_name)

    @classmethod
    def load_server_aws_location(cls, settings: ShieldSettings) -> ShieldedServerLocation:
        location_type: ShieldedServerLocationType
        location_value: str
        if settings.aws_miner_instance_id:
            location_type = ShieldedServerLocationType.EC2_ID
            location_value = settings.aws_miner_instance_id
        elif settings.aws_miner_instance_ip:
            location_type = ShieldedServerLocationType.EC2_IP
            location_value = settings.aws_miner_instance_ip
        else:
            raise MinerShieldException('AWS_MINER_INSTANCE_ID or AWS_MINER_INSTANCE_IP env is not set')
        return ShieldedServerLocation(
            location_type=location_type, location_value=location_value, port=settings.miner_instance_port
        )

    @classmethod
    def create_address_manager(
        cls,
        settings: ShieldSettings,
        aws_client_factory: AWSClientFactory | None,
        event_processor: AbstractMinerShieldEventProcessor,
        state_manager: AbstractMinerShieldStateManager,
    ) -> AbstractAddressManager:
        if aws_client_factory:
            return cls.create_aws_address_manager(settings, aws_client_factory, event_processor, state_manager)
        else:
            raise MinerShieldException('Cannot create address manager')

    @classmethod
    def create_aws_address_manager(
        cls,
        settings: ShieldSettings,
        aws_client_factory: AWSClientFactory,
        event_processor: AbstractMinerShieldEventProcessor,
        state_manager: AbstractMinerShieldStateManager,
    ) -> AbstractAddressManager:
        server_location: ShieldedServerLocation = cls.load_server_aws_location(settings)
        return AwsAddressManager(
            aws_client_factory, server_location, settings.aws_route53_hosted_zone_id, event_processor, state_manager
        )

    @classmethod
    def create_encryption_manager(cls) -> AbstractEncryptionManager:
        return ECIESEncryptionManager()

    @classmethod
    def create_manifest_manager(
        cls,
        settings: ShieldSettings,
        encryption_manager: AbstractEncryptionManager,
        aws_client_factory: AWSClientFactory,
        event_processor: AbstractMinerShieldEventProcessor,
    ) -> AbstractManifestManager:
        manifest_serializer: AbstractManifestSerializer = JsonManifestSerializer()
        return S3ManifestManager(
            manifest_serializer, encryption_manager, event_processor, aws_client_factory, settings.aws_s3_bucket_name
        )

    @classmethod
    def create_blockchain_manager(
        cls,
        settings: ShieldSettings,
        event_processor: AbstractMinerShieldEventProcessor,
    ) -> AbstractBlockchainManager:
        return BittensorBlockchainManager(
            netuid=settings.netuid,
            subtensor=settings.subtensor.client,
            wallet=settings.wallet.instance,
            event_processor=event_processor,
        )


def run_shield() -> int:
    logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')
    parser = argparse.ArgumentParser(description='MinerShield')
    subparsers = parser.add_subparsers(dest='command', help='Subcommands')
    subparsers.add_parser('start', help='Start the MinerShield')
    subparsers.add_parser('clean', help='Clean all stuff created by shield, especially AWS objects')
    args = parser.parse_args()

    settings: ShieldSettings = ShieldSettings()  # type: ignore
    miner_shield: MinerShield = MinerShieldFactory.create_miner_shield(settings)

    if args.command == 'clean':
        logging.info('Cleaning shield objects')
        miner_shield.address_manager.clean_all()
        logging.info('All objects cleaned')
        return 0

    if args.command == 'start' or args.command is None:
        try:
            logging.info('Starting shield')
            miner_shield.enable()
            logging.info('Shield started, press Ctrl+C to stop')
            threading.Event().wait()
            return -1
        except KeyboardInterrupt:
            logging.info('Keyboard interrupt, stopping shield')
            miner_shield.disable()
            return 0
        except MinerShieldException:
            logging.exception('Error during enabling shield')
            return 1

    return -1


if __name__ == '__main__':
    sys.exit(run_shield())



from __future__ import annotations

import asyncio
import time
from dataclasses import dataclass
from typing import TYPE_CHECKING

from bittensor.core.metagraph import Metagraph

from bt_ddos_shield.blockchain_manager import (
    AbstractBlockchainManager,
    BittensorBlockchainManager,
    BlockchainManagerException,
)
from bt_ddos_shield.encryption_manager import AbstractEncryptionManager, ECIESEncryptionManager, EncryptionCertificate
from bt_ddos_shield.event_processor import AbstractMinerShieldEventProcessor, PrintingMinerShieldEventProcessor
from bt_ddos_shield.manifest_manager import (
    JsonManifestSerializer,
    Manifest,
    ManifestDeserializationException,
    ReadOnlyManifestManager,
)

if TYPE_CHECKING:
    import bittensor
    import bittensor_wallet
    from coincurve.keys import PrivateKey as CoincurvePrivateKey

    from bt_ddos_shield.utils import Hotkey, PublicKey


@dataclass
class ShieldMetagraphOptions:
    replace_ip_address_for_axon: bool = True
    """
    Determines how shield address is added to axon info in metagraph. If True, shield address will replace original one
    in `ip` field, otherwise new field `shield_address` will be added. In both cases, port will be replaced with port
    from shield.
    """


class ShieldMetagraph(Metagraph):
    """
    Wrapper class for Metagraph. It allows Validator to retrieve addresses generated by MinerShield instead of Miners
    addresses presented in original metagraph. If given Miner is not shielded, it will return their original address.

    To use this class in your code just replace your Metagraph instance with ShieldMetagraph instance. If you were
    using subtensor.metagraph() before, you should now use constructor with sync param set to True and other params
    set appropriately.

    certificate_path argument is path to file where certificate is stored. If file does not exist, new certificate
    will be generated and saved to this file. If file exists, certificate will be loaded from it.
    """

    wallet: bittensor_wallet.Wallet
    """ Validator's wallet. """
    certificate: EncryptionCertificate
    """ Certificate used for encryption of addresses generated for Validator by Miners. """
    event_processor: AbstractMinerShieldEventProcessor

    encryption_manager: AbstractEncryptionManager
    blockchain_manager: AbstractBlockchainManager
    manifest_manager: ReadOnlyManifestManager
    options: ShieldMetagraphOptions

    def __init__(
        self,
        wallet: bittensor_wallet.Wallet,
        certificate_path: str,
        netuid: int,
        network: str | None = None,
        lite: bool = True,
        sync: bool = True,
        block: int | None = None,
        subtensor: bittensor.Subtensor | None = None,
        event_processor: AbstractMinerShieldEventProcessor | None = None,
        encryption_manager: AbstractEncryptionManager | None = None,
        blockchain_manager: AbstractBlockchainManager | None = None,
        manifest_manager: ReadOnlyManifestManager | None = None,
        options: ShieldMetagraphOptions | None = None,
    ):
        super().__init__(
            netuid=netuid,
            lite=lite,
            sync=False,
            subtensor=subtensor,
            **({'network': network} if network is not None else {}),
        )

        self.wallet = wallet
        self.options = options or ShieldMetagraphOptions()
        self.event_processor = event_processor or PrintingMinerShieldEventProcessor()
        self.encryption_manager = encryption_manager or self.create_default_encryption_manager()
        self.blockchain_manager = blockchain_manager or self.create_default_blockchain_manager(
            subtensor, netuid, wallet, self.event_processor
        )
        self.manifest_manager = manifest_manager or self.create_default_manifest_manager(
            self.event_processor, self.encryption_manager
        )
        self._init_certificate(certificate_path)

        if sync:
            self.sync(block=block, lite=lite, subtensor=subtensor)
        elif block is not None:
            raise ValueError('Block argument is valid only when sync is True')

    def _init_certificate(self, certificate_path: str):
        try:
            coincurve_cert: CoincurvePrivateKey = self.encryption_manager.load_certificate(certificate_path)
            self.certificate = self.encryption_manager.serialize_certificate(coincurve_cert)
            public_key: PublicKey | None = self.blockchain_manager.get_own_public_key()
            if self.certificate.public_key == public_key:
                return
        except FileNotFoundError:
            coincurve_cert = self.encryption_manager.generate_certificate()
            self.encryption_manager.save_certificate(coincurve_cert, certificate_path)
            self.certificate = self.encryption_manager.serialize_certificate(coincurve_cert)

        try:
            self.blockchain_manager.upload_public_key(self.certificate.public_key)
        except BlockchainManagerException:
            # Retry once
            time.sleep(3)
            self.blockchain_manager.upload_public_key(self.certificate.public_key)

    @classmethod
    def create_default_encryption_manager(cls):
        return ECIESEncryptionManager()

    @classmethod
    def create_default_manifest_manager(
        cls,
        event_processor: AbstractMinerShieldEventProcessor,
        encryption_manager: AbstractEncryptionManager,
    ) -> ReadOnlyManifestManager:
        return ReadOnlyManifestManager(JsonManifestSerializer(), encryption_manager, event_processor)

    @classmethod
    def create_default_blockchain_manager(
        cls,
        subtensor: bittensor.Subtensor,
        netuid: int,
        wallet: bittensor_wallet.Wallet,
        event_processor: AbstractMinerShieldEventProcessor,
    ) -> AbstractBlockchainManager:
        return BittensorBlockchainManager(
            subtensor=subtensor,
            netuid=netuid,
            wallet=wallet,
            event_processor=event_processor,
        )

    def sync(self, block: int | None = None, lite: bool = True, subtensor: bittensor.Subtensor | None = None):
        super().sync(block=block, lite=lite, subtensor=subtensor)
        hotkeys: list[str] = self.hotkeys
        urls: dict[Hotkey, str | None] = asyncio.run(self.blockchain_manager.get_manifest_urls(hotkeys))
        manifests: dict[Hotkey, Manifest | None] = asyncio.run(self.manifest_manager.get_manifests(urls))
        own_hotkey: Hotkey = self.wallet.hotkey.ss58_address
        for axon in self.axons:
            manifest: Manifest | None = manifests.get(axon.hotkey)
            if manifest is not None:
                try:
                    shield_address: tuple[str, int] | None = self.manifest_manager.get_address_for_validator(
                        manifest, own_hotkey, self.certificate.private_key
                    )
                except ManifestDeserializationException as e:
                    self.event_processor.event(
                        'Error while getting shield address for miner {hotkey}', exception=e, axon=axon.hotkey
                    )
                    continue
                if shield_address is not None:
                    if self.options.replace_ip_address_for_axon:
                        axon.ip = shield_address[0]
                    else:
                        axon.shield_address = shield_address[0]
                    axon.port = shield_address[1]



from __future__ import annotations

import os
from typing import TYPE_CHECKING

import pytest

from bt_ddos_shield.encryption_manager import (
    DecryptionError,
    ECIESEncryptionManager,
    EncryptionCertificate,
    EncryptionError,
)

if TYPE_CHECKING:
    from coincurve.keys import PrivateKey as CoincurvePrivateKey

# Sample test data
valid_test_data = b'encrypted_address'
non_encrypted_bytes = b'This is not encrypted'
invalid_key = 'invalid_key'


class TestEncryptionManager:
    """
    Test suite for the EncryptionManager class.
    """

    encryption_manager = ECIESEncryptionManager()
    private_key, public_key = encryption_manager.serialize_certificate(encryption_manager.generate_certificate())

    def test_encrypt_data_valid(self):
        """
        Test encryption with valid public key and data.
        """
        self.encryption_manager.encrypt(public_key=self.public_key, data=valid_test_data)

    def test_encrypt_data_invalid_public_key(self):
        """
        Test encryption with an invalid public key (string that doesn't represent a valid key).
        Expects EncryptionError to be raised.
        """
        with pytest.raises(EncryptionError):
            self.encryption_manager.encrypt(public_key=invalid_key, data=valid_test_data)

    def test_decrypt_data_valid(self):
        """
        Test decryption with valid private key and encrypted data.
        Ensures that the decrypted data matches the original one.
        """
        encrypted_data = self.encryption_manager.encrypt(public_key=self.public_key, data=valid_test_data)
        decrypted_data = self.encryption_manager.decrypt(private_key=self.private_key, data=encrypted_data)
        assert decrypted_data == valid_test_data, 'Decrypted data should match the original data'

    def test_decrypt_data_invalid_private_key(self):
        """
        Test decryption with an invalid private key (string that doesn't represent a valid key).
        Expects DecryptionError to be raised.
        """
        encrypted_data = self.encryption_manager.encrypt(public_key=self.public_key, data=valid_test_data)
        with pytest.raises(DecryptionError):
            self.encryption_manager.decrypt(private_key=invalid_key, data=encrypted_data)

    def test_decrypt_data_invalid_encrypted_data(self):
        """
        Test decryption with invalid encrypted data (non-encrypted bytes).
        Expects DecryptionError to be raised.
        """
        with pytest.raises(DecryptionError):
            self.encryption_manager.decrypt(private_key=self.private_key, data=non_encrypted_bytes)

    def test_save_and_load_certificate(self) -> None:
        """
        Test saving and loading a certificate to/from disk.
        """
        path: str = 'certificate_test.pem'
        certificate: CoincurvePrivateKey = self.encryption_manager.generate_certificate()
        try:
            self.encryption_manager.save_certificate(certificate, path)
            loaded_certificate: CoincurvePrivateKey = self.encryption_manager.load_certificate(path)
            assert certificate.to_hex() == loaded_certificate.to_hex()
            serialized_cert: EncryptionCertificate = self.encryption_manager.serialize_certificate(loaded_certificate)
            encrypted_data = self.encryption_manager.encrypt(
                public_key=serialized_cert.public_key, data=valid_test_data
            )
            decrypted_data = self.encryption_manager.decrypt(
                private_key=serialized_cert.private_key, data=encrypted_data
            )
            assert decrypted_data == valid_test_data
        finally:
            os.remove(path)



from __future__ import annotations

import asyncio
import base64
import functools
import hashlib
import json
import time
from abc import ABC, abstractmethod
from dataclasses import asdict, dataclass
from http import HTTPStatus
from typing import TYPE_CHECKING, Any

import aiohttp

if TYPE_CHECKING:
    from types import MappingProxyType

    from mypy_boto3_s3 import S3Client

    from bt_ddos_shield.encryption_manager import AbstractEncryptionManager
    from bt_ddos_shield.event_processor import AbstractMinerShieldEventProcessor
    from bt_ddos_shield.utils import AWSClientFactory, Hotkey, PrivateKey, PublicKey, ShieldAddress


class ManifestManagerException(Exception):
    pass


class ManifestDeserializationException(ManifestManagerException):
    """
    Exception thrown when deserialization of manifest data fails.
    """


class ManifestDownloadException(ManifestManagerException):
    """
    Exception thrown when error occurs during downloading manifest file.
    """


class ManifestNotFoundException(ManifestDownloadException):
    """
    Exception thrown when manifest file is not found under given address.
    """


@dataclass
class Manifest:
    """
    Class representing manifest file containing encrypted addresses for validators.
    """

    encrypted_url_mapping: dict[Hotkey, bytes]
    """ Mapping with addresses for validators (validator HotKey -> encrypted url) """
    md5_hash: str
    """ MD5 hash of the manifest data """


class AbstractManifestSerializer(ABC):
    """
    Class used to serialize and deserialize manifest file.
    """

    @abstractmethod
    def serialize(self, manifest: Manifest) -> bytes:
        """
        Serialize manifest. Output format depends on the implementation.
        """
        pass

    @abstractmethod
    def deserialize(self, serialized_data: bytes) -> Manifest:
        """
        Deserialize manifest. Throws ManifestDeserializationException if data format is not recognized.
        """
        pass


class JsonManifestSerializer(AbstractManifestSerializer):
    """
    Manifest serializer implementation which serialize manifest to Json.
    """

    MANIFEST_ROOT_JSON_KEY: str = 'ddos_shield_manifest'

    encoding: str

    def __init__(self, encoding: str = 'utf-8'):
        """
        Args:
            encoding: Encoding used for transforming Json string to bytes.
        """
        self.encoding = encoding

    def serialize(self, manifest: Manifest) -> bytes:
        data: dict = {
            self.MANIFEST_ROOT_JSON_KEY: asdict(manifest)  # type: ignore
        }
        json_str: str = json.dumps(data, default=self._custom_encoder)
        return json_str.encode(encoding=self.encoding)

    def deserialize(self, serialized_data: bytes) -> Manifest:
        try:
            json_str: str = serialized_data.decode(encoding=self.encoding)
            data = json.loads(json_str, object_hook=self._custom_decoder)
            return Manifest(**data[self.MANIFEST_ROOT_JSON_KEY])
        except Exception as e:
            raise ManifestDeserializationException(f'Failed to deserialize manifest data: {e}') from e

    @staticmethod
    def _custom_encoder(obj: Any) -> Any:
        if isinstance(obj, bytes):
            return base64.b64encode(obj).decode()  # type: ignore

    @staticmethod
    def _custom_decoder(json_mapping: dict[str, Any]) -> Any:
        if 'encrypted_url_mapping' in json_mapping:
            decoded_mapping: dict[Hotkey, bytes] = {}
            for hotkey, encoded_address in json_mapping['encrypted_url_mapping'].items():
                decoded_mapping[hotkey] = base64.b64decode(encoded_address.encode())
            json_mapping['encrypted_url_mapping'] = decoded_mapping
        return json_mapping


class ReadOnlyManifestManager(ABC):
    """
    Manifest manager only for getting file uploaded by ManifestManager.
    """

    manifest_serializer: AbstractManifestSerializer
    encryption_manager: AbstractEncryptionManager
    event_processor: AbstractMinerShieldEventProcessor
    _download_timeout: int

    def __init__(
        self,
        manifest_serializer: AbstractManifestSerializer,
        encryption_manager: AbstractEncryptionManager,
        event_processor: AbstractMinerShieldEventProcessor,
        download_timeout: int = 10,
    ):
        self.manifest_serializer = manifest_serializer
        self.encryption_manager = encryption_manager
        self.event_processor = event_processor
        self._download_timeout = download_timeout

    async def get_manifest(self, url: str) -> Manifest:
        """
        Get manifest file from given url and deserialize it.
        Throws ManifestNotFoundException if file is not found.
        Throws ManifestDeserializationException if data format is not recognized.
        """
        async with aiohttp.ClientSession(timeout=aiohttp.ClientTimeout(total=self._download_timeout)) as session:
            raw_data: bytes | None = await self._get_manifest_file(session, url)
        if raw_data is None:
            raise ManifestDownloadException(f'Manifest file not found at {url}')
        return self.manifest_serializer.deserialize(raw_data)

    async def get_manifests(self, urls: dict[Hotkey, str | None]) -> dict[Hotkey, Manifest | None]:
        """
        Get manifest files from given urls and deserialize them. If url is None, None is returned in the result.
        None is also returned if manifest file is not found or has invalid format.

        Args:
            urls: Dictionary with urls for neurons (neuron HotKey -> url).
        """
        raw_manifests_data: dict[Hotkey, bytes | None] = await self._get_manifest_files(urls)
        manifests: dict[Hotkey, Manifest | None] = {}
        for hotkey, raw_data in raw_manifests_data.items():
            manifest: Manifest | None = None
            if raw_data is not None:
                try:
                    manifest = self.manifest_serializer.deserialize(raw_data)
                except ManifestDeserializationException:
                    self.event_processor.event(
                        'Manifest file corrupted for hotkey={hotkey}, url={url}', hotkey=hotkey, url=urls[hotkey]
                    )
            manifests[hotkey] = manifest
        return manifests

    def get_address_for_validator(
        self, manifest: Manifest, validator_hotkey: Hotkey, validator_private_key: PrivateKey
    ) -> tuple[str, int] | None:
        """
        Get URL and port for validator identified by hotkey from manifest or None if not found.
        Decrypts address using validator's private key.
        Throws ManifestDeserializationException if address format is invalid.
        """
        encrypted_url: bytes | None = manifest.encrypted_url_mapping.get(validator_hotkey)
        if encrypted_url is None:
            return None
        try:
            decrypted_url: bytes = self.encryption_manager.decrypt(validator_private_key, encrypted_url)
            url: str = decrypted_url.decode()
            parts: list[str] = url.split(':')
            return parts[0], int(parts[1])
        except Exception as e:
            raise ManifestDeserializationException(
                f'Invalid address format for validator {validator_hotkey}: {e}'
            ) from e

    async def _get_manifest_file(self, http_session: aiohttp.ClientSession, url: str | None) -> bytes | None:
        if url is None:
            return None

        try:
            async with http_session.get(url) as response:
                response.raise_for_status()
                return await response.read()
        except aiohttp.ClientResponseError as e:
            if e.status in (HTTPStatus.FORBIDDEN, HTTPStatus.NOT_FOUND):
                # REMARK: S3 returns 403 Forbidden if file does not exist in bucket.
                self.event_processor.event(
                    'Manifest file not found, url={url}, status code={status_code}', url=url, status_code=e.status
                )
                return None
            raise ManifestDownloadException(f'HTTP error when downloading file from {url}: {e}') from e
        except (TimeoutError, aiohttp.ClientConnectionError) as e:
            raise ManifestDownloadException(f'Failed to download file from {url}: {e}') from e

    async def _get_manifest_file_with_retry(self, http_session: aiohttp.ClientSession, url: str | None) -> bytes | None:
        try:
            return await self._get_manifest_file(http_session, url)
        except ManifestDownloadException:
            # Retry once
            time.sleep(1)
            return await self._get_manifest_file(http_session, url)

    async def _get_manifest_files(self, urls: dict[Hotkey, str | None]) -> dict[Hotkey, bytes | None]:
        """
        Get manifest files from given urls. If url is None, None is returned in the result. None is also returned if
        manifest file is not found.

        Args:
            urls: Dictionary with urls for neurons (neuron HotKey -> url).
        """
        async with aiohttp.ClientSession(timeout=aiohttp.ClientTimeout(total=self._download_timeout)) as session:
            tasks = [self._get_manifest_file_with_retry(session, url) for url in urls.values()]
            results: list[bytes | None] = await asyncio.gather(*tasks)
        return dict(zip(urls.keys(), results, strict=True))


class AbstractManifestManager(ReadOnlyManifestManager):
    """
    Abstract base class for manager handling manifest file containing encrypted addresses for validators.
    """

    def upload_manifest(self, manifest: Manifest):
        data: bytes = self.manifest_serializer.serialize(manifest)
        return self._put_manifest_file(data)

    def create_manifest(
        self,
        address_mapping: MappingProxyType[Hotkey, ShieldAddress],
        validators_public_keys: MappingProxyType[Hotkey, PublicKey],
    ) -> Manifest:
        """
        Create manifest with encrypted addresses for validators.

        Args:
            address_mapping: Dictionary containing address mapping (validator HotKey -> Address).
            validators_public_keys: Dictionary containing public keys of validators (validator HotKey -> PublicKey).
        """
        encrypted_address_mapping: dict[Hotkey, bytes] = {}
        md5_hash = hashlib.md5(usedforsecurity=False)

        for hotkey, address in address_mapping.items():
            public_key: PublicKey = validators_public_keys[hotkey]
            url: str = f'{address.address}:{address.port}'
            serialized_url: bytes = url.encode()
            encrypted_address_mapping[hotkey] = self.encryption_manager.encrypt(public_key, serialized_url)

            md5_hash.update(hotkey.encode())  # type: ignore
            public_key_bytes: bytes = public_key.encode() if isinstance(public_key, str) else public_key
            md5_hash.update(public_key_bytes)  # type: ignore
            md5_hash.update(serialized_url)  # type: ignore

        return Manifest(encrypted_address_mapping, md5_hash.hexdigest())

    @abstractmethod
    def get_manifest_url(self) -> str:
        """
        Return URL where manifest file is stored.
        """
        pass

    @abstractmethod
    def _put_manifest_file(self, data: bytes):
        """
        Put manifest file into the storage. Should overwrite manifest file if it exists.
        """
        pass


class S3ManifestManager(AbstractManifestManager):
    """
    Manifest manager using AWS S3 service to manage file.
    """

    MANIFEST_FILE_NAME: str = 'shield_manifest.json'

    _aws_client_factory: AWSClientFactory
    _bucket_name: str

    def __init__(
        self,
        manifest_serializer: AbstractManifestSerializer,
        encryption_manager: AbstractEncryptionManager,
        event_processor: AbstractMinerShieldEventProcessor,
        aws_client_factory: AWSClientFactory,
        bucket_name: str,
        download_timeout: int = 10,
    ):
        super().__init__(manifest_serializer, encryption_manager, event_processor, download_timeout)
        self._aws_client_factory = aws_client_factory
        self._bucket_name = bucket_name

    def get_manifest_url(self) -> str:
        assert self._aws_client_factory.aws_region_name is not None, 'aws_region_name needs to be initialized first'
        region_name: str = self._aws_client_factory.aws_region_name
        return f'https://{self._bucket_name}.s3.{region_name}.amazonaws.com/{self.MANIFEST_FILE_NAME}'

    @functools.cached_property
    def _s3_client(self) -> S3Client:
        return self._aws_client_factory.boto3_client('s3')  # type: ignore

    def _put_manifest_file(self, data: bytes):
        self._s3_client.put_object(Bucket=self._bucket_name, Key=self.MANIFEST_FILE_NAME, Body=data, ACL='public-read')



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING

import aiohttp

from bt_ddos_shield.encryption_manager import ECIESEncryptionManager
from bt_ddos_shield.event_processor import PrintingMinerShieldEventProcessor
from bt_ddos_shield.manifest_manager import (
    AbstractManifestManager,
    JsonManifestSerializer,
    Manifest,
    ReadOnlyManifestManager,
    S3ManifestManager,
)
from bt_ddos_shield.utils import AWSClientFactory, Hotkey

if TYPE_CHECKING:
    from tests.conftest import ShieldTestSettings


class MemoryManifestManager(AbstractManifestManager):
    _manifest_url: str
    stored_file: bytes | None
    put_counter: int

    def __init__(self):
        super().__init__(JsonManifestSerializer(), ECIESEncryptionManager(), PrintingMinerShieldEventProcessor())
        self._manifest_url = 'https://manifest.com'
        self.stored_file = None
        self.put_counter = 0

    def get_manifest_url(self) -> str:
        return self._manifest_url

    def _put_manifest_file(self, data: bytes):
        self.stored_file = data
        self.put_counter += 1

    async def _get_manifest_file(self, http_session: aiohttp.ClientSession, url: str | None) -> bytes | None:
        if self.stored_file is None or url != self._manifest_url:
            return None
        return self.stored_file


class TestManifestManager:
    """
    Test suite for the manifest manager.
    """

    def test_json_serializer(self) -> None:
        manifest_serializer = JsonManifestSerializer()
        mapping: dict[Hotkey, bytes] = {'validator1': b'address1', 'validator2': b'address2'}
        md5_hash: str = 'some_hash'
        manifest: Manifest = Manifest(mapping, md5_hash)
        json_data: bytes = manifest_serializer.serialize(manifest)
        deserialized_manifest: Manifest = manifest_serializer.deserialize(json_data)
        assert manifest == deserialized_manifest

    def test_s3_put_get(self, shield_settings: ShieldTestSettings):
        """Test S3ManifestManager class. Put manifest file, get it and check if it was stored correctly."""
        aws_client_factory: AWSClientFactory = AWSClientFactory(
            shield_settings.aws_access_key_id, shield_settings.aws_secret_access_key, shield_settings.aws_region_name
        )
        manifest_manager = S3ManifestManager(
            aws_client_factory=aws_client_factory,
            bucket_name=shield_settings.aws_s3_bucket_name,
            manifest_serializer=JsonManifestSerializer(),
            encryption_manager=ECIESEncryptionManager(),
            event_processor=PrintingMinerShieldEventProcessor(),
        )
        http_session: aiohttp.ClientSession = aiohttp.ClientSession(loop=asyncio.get_event_loop())

        try:
            data: bytes = b'some_data'
            manifest_manager._put_manifest_file(data)
            manifest_url: str = manifest_manager.get_manifest_url()
            retrieved_data: bytes | None = asyncio.run(manifest_manager._get_manifest_file(http_session, manifest_url))
            assert retrieved_data == data
            assert asyncio.run(manifest_manager._get_manifest_file(http_session, manifest_url + 'xxx')) is None

            other_data: bytes = b'other_data'
            manifest_manager._put_manifest_file(other_data)
            retrieved_data = asyncio.run(manifest_manager._get_manifest_file(http_session, manifest_url))
            assert retrieved_data == other_data

            validator_manifest_manager = ReadOnlyManifestManager(
                manifest_serializer=JsonManifestSerializer(),
                encryption_manager=ECIESEncryptionManager(),
                event_processor=PrintingMinerShieldEventProcessor(),
            )
            retrieved_data = asyncio.run(validator_manifest_manager._get_manifest_file(http_session, manifest_url))
            assert retrieved_data == other_data
        finally:
            asyncio.run(http_session.close())



from __future__ import annotations

import threading
from typing import TYPE_CHECKING

from bt_ddos_shield.blockchain_manager import (
    AbstractBlockchainManager,
)

if TYPE_CHECKING:
    from collections.abc import Iterable

    from bt_ddos_shield.utils import Hotkey, PublicKey


class MemoryBlockchainManager(AbstractBlockchainManager):
    known_data: dict[Hotkey, bytes]
    put_counter: int

    def __init__(self, miner_hotkey: Hotkey):
        self.miner_hotkey = miner_hotkey
        self.known_data = {}
        self.put_counter = 0
        self._lock = threading.Lock()

    def get_hotkey(self) -> Hotkey:
        return self.miner_hotkey

    def put_metadata(self, data: bytes):
        with self._lock:
            self.known_data[self.miner_hotkey] = data
            self.put_counter += 1

    async def get_metadata(self, hotkeys: Iterable[Hotkey]) -> dict[Hotkey, bytes | None]:
        with self._lock:
            return {hotkey: self.known_data.get(hotkey) for hotkey in hotkeys}

    def get_own_public_key(self) -> PublicKey:
        return 'public_key'

    def upload_public_key(self, public_key: PublicKey):
        pass



from __future__ import annotations

import functools
from dataclasses import dataclass
from typing import TYPE_CHECKING

import bittensor
import boto3
import route53
from pydantic import BaseModel

if TYPE_CHECKING:
    import bittensor_wallet
    from botocore.client import BaseClient
    from route53.connection import Route53Connection

type Hotkey = str
type PublicKey = str
type PrivateKey = str


@dataclass
class ShieldAddress:
    """
    Class describing address created by DDosShield.
    """

    address_id: str
    """ Identifier of the address """
    address: str
    """ Domain address used to connecting to Miner's server """
    port: int
    """ Port used to connecting to Miner's server """

    def __repr__(self):
        return f'Address(id={self.address_id}, address={self.address}:{self.port})'


class AWSClientFactory:
    aws_access_key_id: str
    aws_secret_access_key: str
    aws_region_name: str | None

    def __init__(self, aws_access_key_id: str, aws_secret_access_key: str, aws_region_name: str | None = None):
        """
        Args:
            aws_access_key_id: AWS access key ID.
            aws_secret_access_key: AWS secret access key.
            aws_region_name: AWS region name. If not known, it can be set later using set_aws_region_name method.
        """
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.aws_region_name = aws_region_name or ''

    def set_aws_region_name(self, aws_region_name: str) -> bool:
        """Set AWS region name. Returns if region name was changed."""
        if self.aws_region_name == aws_region_name:
            return False
        self.aws_region_name = aws_region_name
        return True

    def boto3_client(self, service_name: str) -> BaseClient:
        return boto3.client(  # type: ignore
            service_name,
            aws_access_key_id=self.aws_access_key_id,
            aws_secret_access_key=self.aws_secret_access_key,
            region_name=self.aws_region_name,
        )

    def route53_client(self) -> Route53Connection:
        return route53.connect(self.aws_access_key_id, self.aws_secret_access_key)


class WalletSettings(BaseModel):
    name: str | None = None
    hotkey: str | None = None
    path: str | None = None

    @functools.cached_property
    def instance(self) -> bittensor_wallet.Wallet:
        return bittensor.Wallet(**self.model_dump())


class SubtensorSettings(BaseModel):
    network: str | None = None

    @functools.cached_property
    def client(self) -> bittensor.Subtensor:
        return self.create_client()

    def create_client(self) -> bittensor.Subtensor:
        return bittensor.Subtensor(**self.model_dump())
